/*

        Comma separated value database format scanner

        This function implements a somewhat extended flavour
        of CSV.  In addition to the standard quoted fields,
        permitting embedded commas, with embedded quotes
        represented as "", backslash escaped character
        expressed as three octal digits are also permitted,
        with a double backslash representing an embedded
        backslash.  This is necessary to permit fields
        which include end-of-line delimiters which would
        otherwise truncate the record when it is read.

*/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "csv.h"

#define EOS     '\0'

#define false   0
#define true    1

#define isOctal(c)  (((c) >= '0') && ((c) <= '7'))

static char *csptr;                   /* CSV scan pointer */

/*  CSVSCANINIT  --  Initialise scanning of a CSV record.  */

void CSVscanInit(char *s)
{
    csptr = s;
}

/*  CSVSCANFIELD  --  Scan next field from a CSV record.
                      Returns true if field found, false at
                      end of record.  */

int CSVscanField(char *f)
{
    char *ef = f;
    int foundfield = false, quoted = false;

    if (*csptr != EOS) {
        foundfield = true;
        while ((*csptr != EOS) && isspace(*csptr)) {
            csptr++;
        }
        if (*csptr == '"') {
            quoted = true;
            csptr++;
            while (*csptr != EOS) {
                if (*csptr == '"') {
                    if (csptr[1] == '"') {
                        *f++ = '"';
                        csptr += 2;
                    } else {
                        csptr++;
                        break;
                    }
                } else if (*csptr == '\\') {
                    if (csptr[1] == '\\') {
                        *f++ = '\\';
                        csptr += 2;
                    } else if (isOctal(csptr[1]) && isOctal(csptr[2]) && isOctal(csptr[3])) {
                        unsigned int v;

                        v = ((csptr[1] - '0') << 6) | ((csptr[2] - '0') << 3) | (csptr[3] - '0');
                        *f++ = (char) v;
                        csptr += 4;
                    } else {
                        *f++ = csptr[1];
                        csptr += 2;
                    }
                } else {
                    *f++ = *csptr++;
                }
            }
        }
        while (*csptr != ',' && *csptr != EOS) {
            *f++ = *csptr++;
        }
        if (*csptr == ',') {
            csptr++;
        }
    }
    *f++ = EOS;
    if (foundfield && !quoted) {
        while (strlen(ef) > 0 && isspace(ef[strlen(ef) - 1])) {
            ef[strlen(ef) - 1] = EOS;
        }
    }
    return foundfield;
}
